@@ -13,6 +13,7 @@
 #include <signal.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
+#include <errno.h>
 
 
 #ifdef __sgi
@@ -52,23 +53,35 @@ static char nfshost[128];
 # endif
 #endif /* __FreeBSD__ */
 
+#ifdef linux
+#define _LINUX_IN_H   1
+#endif
+
 #include <rpc/rpc.h>
+#ifdef linux
+#define _LINUX_SUNRPC_MSGPROT_H_
+#define __KERNEL__		/* define nfs_fh for nfs_mount_data */
 #include "nfs_prot.h"
+#undef __KERNEL__
+#else
+#include "nfs_prot.h"
+#endif
 
 #ifdef linux
 # ifndef __GLIBC__
 #   include <linux/fs.h>	/* struct nfs_mount_data */
 # endif
+# define _LINUX_NFS2_H		/* Do not include nfs2.h, we have it already */
 # include <linux/nfs_mount.h>	/* struct nfs_mount_data */
 # include <arpa/inet.h>		/* inet_addr() */
 #endif
+
 #ifdef _IBMR2
 # include "os-aix3.h"
 # include "misc-aix3.h"
 # include <sys/vmount.h>
 #endif
 
-
 #ifndef DONT_UPDATE_MTAB
 #if defined(sun) && defined(__SVR4)
 #include <sys/mnttab.h>
@@ -164,7 +177,8 @@ doexit()
 #else
   if(umount(mntdir))
     {
-      perror(mntdir);
+      fprintf(stderr, "umount %s:", mntdir);
+      perror("");
       return;
     }
 #endif
@@ -207,9 +221,12 @@ doexit()
 #else
   if(debug) printf("no mtab fixing needed\n");
 #endif
+
   reset_serial(psionfd);
+  close(psionfd);
+
   fprintf(stderr, "p3nfsd: exiting.\n");
-  
+
   exit(0);
 }
 
@@ -230,7 +247,7 @@ dosystem(str)
 void
 mount_and_run(dir, dev, proc, root_fh)
   char *dir, *dev;
-  nfs_fh *root_fh;
+  mynfs_fh *root_fh;
   void (*proc)();
 {
   int sock, port, pid, doclean;
@@ -257,7 +274,7 @@ mount_and_run(dir, dev, proc, root_fh)
 
 
 
-  sprintf(nfshost, "localhost:%s", dev);
+  sprintf(nfshost, "localhost:%s", dev ? dev : "tcp");
   bzero((char *)&sain, sizeof(struct sockaddr_in));
 #ifdef linux
   bzero(&nfs_mount_data, sizeof(struct nfs_mount_data));
@@ -294,7 +311,7 @@ mount_and_run(dir, dev, proc, root_fh)
     }
   if (debug)
     printf("created svc PROG=%d, VER=%d port=%d\n",
-    	NFS_PROGRAM, NFS_VERSION, port);
+    	NFS_PROGRAM, (int)NFS_VERSION, port);
     
 
 /*** Second part: mount the directory */
@@ -325,9 +342,14 @@ mount_and_run(dir, dev, proc, root_fh)
 
   nfs_mount_data.version = NFS_MOUNT_VERSION;
   nfs_mount_data.fd      = ksock;
+#if NFS_MOUNT_VERSION >= 4
+  nfs_mount_data.root.size = sizeof(*root_fh);
+  memcpy(nfs_mount_data.root.data, root_fh, sizeof(*root_fh));
+#else
   nfs_mount_data.root    = *root_fh;	/* structure copy */
+#endif
   nfs_mount_data.flags   = NFS_MOUNT_INTR | NFS_MOUNT_NOCTO; 
-  			/* NFS_MOUNT_SECURE | NFS_MOUNT_POSIX | NFS_MOUNT_SOFT | NFS_MOUNT_NOAC */
+  			/* NFS_MOUNT_SECURE | NFS_MOUNT_SOFT | NFS_MOUNT_NOAC */
   nfs_mount_data.rsize    = PBUFSIZE;
   nfs_mount_data.wsize    = PBUFSIZE;
   nfs_mount_data.timeo    = 600;
@@ -356,7 +378,6 @@ mount_and_run(dir, dev, proc, root_fh)
     struct knetconfig knc;
     struct netconfig *nc=getnetconfigent("udp");
     if (!nc) {
-	extern int errno;
 	fprintf(stderr, "getnetconfigent \"udp\": (errno %d) ", errno);
 	perror("");
 	exit(1);
@@ -366,7 +387,6 @@ mount_and_run(dir, dev, proc, root_fh)
     knc.knc_proto=strdup(nc->nc_proto);
 
     if (stat(nc->nc_device,&stb)) {
-	extern int errno;
 	fprintf(stderr, "stat \"%s\": (errno %d) ",nc->nc_device, errno);
 	perror("");
 	exit(1);
@@ -510,8 +530,6 @@ mount_and_run(dir, dev, proc, root_fh)
 #endif
 
 	  {
-	    extern int errno;
-
 	    fprintf(stderr, "nfs mount %s: (errno %d) ", dir, errno);
 	    perror("");
 	    kill(pid, SIGTERM);
@@ -561,13 +579,7 @@ mount_and_run(dir, dev, proc, root_fh)
   printf("p3nfsd: to stop the server do \"ls %s/exit\". (pid %d)\n",
          mntdir, (int)getpid());
 
-#if defined(sun) && !defined(__SVR4)
-  dtbsize = _rpc_dtablesize();
-  if(psionfd  >= dtbsize) dtbsize = psionfd+1;
-  if(masterfd >= dtbsize) dtbsize = masterfd+1;
-#else
   dtbsize = FD_SETSIZE;
-#endif
 
   /*
    * Signal handling, etc. should only happen in one process, so let's
@@ -592,21 +604,28 @@ mount_and_run(dir, dev, proc, root_fh)
       ret = select(dtbsize, &readfd, 0, 0, &tv);
       if(ret > 0)
         {
-	  if(FD_ISSET(psionfd, &readfd)) /* Terminal input from the psion */
+	  if(psionfd >= 0 &&
+	     FD_ISSET(psionfd, &readfd)) /* Terminal input from the psion */
 	    {
 	      unsigned char c;
 
 	      /* Make sure the OS hasn't closed psionfd on me! */
-	      if(read(psionfd, &c, 1) <= 0)
+	      int len = read(psionfd, &c, 1);
+	      if(len <= 0)
 	        {
+	          if(debug)
+		  	printf("p3nfsd: read psionfd: len:%d, errno:%d\n",
+		  			len, errno);
 		  close(psionfd);
 		  psionfd = -1;
+		  doexit();
 		  continue;
 		}
 	      shell_feed(c);
 	      continue;
 	    }
-	  if(masterfd >=0 && FD_ISSET(masterfd, &readfd))
+	  if(masterfd >=0 &&
+	     FD_ISSET(masterfd, &readfd))
 	    /* Terminal output */
 	    {
 #define PLEN 200		/* to make life easier for the psion */
