@@ -1,6 +1,7 @@
 #include "OSdefs.h"
 #include <stdio.h>
 #include <ctype.h>
+#include <time.h>
 #if defined(__SVR4) || defined(__GLIBC__)
 #include <stdlib.h>
 #include <unistd.h>
@@ -25,14 +26,17 @@ create_it(ca, isdir)
   int isdir;
 {
   static struct diropres res;
-  p_inode *dirinode = get_num(fh2inode(ca->where.dir.data));
-  char *name = dirinode->name;
+  p_inode *dirinode;
+  char *name;
   fattr *fp;
   p_inode *inode;
 
   if(debug)
-    printf("\tcreate: in %s %s (%#o, %d)\n",
-	    name, ca->where.name, ca->attributes.mode, isdir);
+    printf("\tcreate: %s (%#o,%d)\n",ca->where.name,ca->attributes.mode,isdir);
+  dirinode = get_num(fh2inode(ca->where.dir.data));
+  name = dirinode->name;
+  if(debug)
+    printf("\tcreate: (in %s)\n", name);
 
   name = build_path(name, ca->where.name);
 
@@ -185,15 +189,17 @@ query_devices()
 
 struct attrstat *
 nfsproc_getattr_2(fh)
-  struct nfs_fh *fh;
+  struct mynfs_fh *fh;
 {
   static struct attrstat res;
-  p_inode *inode = get_num(fh2inode(fh->data));
+  p_inode *inode;
   fattr *fp = &res.attrstat_u.attributes;
   unsigned char op[12];
   struct cache *cp;
   int l;
 
+  if(debug > 2) printf("\tgetattr\n");
+  inode = get_num(fh2inode(fh->data));
   if(debug) printf("\tgetattr:'%s',%d\n", inode->name, inode->inode);
   res.status = NFS_OK;
 
@@ -277,9 +283,11 @@ nfsproc_lookup_2(da)
 {
   static struct diropres res;
   struct attrstat *gres;
-  p_inode *inode = get_num(fh2inode(da->dir.data));
+  p_inode *inode;
   char *fp = res.diropres_u.diropres.file.data;
 
+  if(debug > 2) printf("lookup\n");
+  inode = get_num(fh2inode(da->dir.data));
   if(!inode)
     {
       if(debug) printf("lookup: stale fh\n");
@@ -334,7 +342,7 @@ nfsproc_lookup_2(da)
   else
     inode2fh(get_nam(build_path(inode->name, da->name))->inode, fp);
 
-  gres = nfsproc_getattr_2((struct nfs_fh *)fp);
+  gres = nfsproc_getattr_2((struct mynfs_fh *)fp);
 
   res.status = gres->status;
   res.diropres_u.diropres.attributes = gres->attrstat_u.attributes;
@@ -399,6 +407,7 @@ addentry(ra, where, searchinode, inode, name)
   *(int *)(**where)->cookie = inode;
   (**where)->nextentry = 0;
   *where = &(**where)->nextentry;
+
 }
 
 struct readdirres *
@@ -406,11 +415,13 @@ nfsproc_readdir_2(ra)
   readdirargs *ra;
 {
   static readdirres res;
-  p_inode *inode = get_num(fh2inode(ra->dir.data));
+  p_inode *inode;
   entry *centry, *fentry, **cp;
   char *sp, *p, *bp, buf[256];
   int searchinode;
 
+  if(debug > 2) printf("readdir\n");
+  inode = get_num(fh2inode(ra->dir.data));
   if(!inode)
     {
       if(debug) printf("readdir: stale fh\n");
@@ -431,6 +442,8 @@ nfsproc_readdir_2(ra)
 	   inode->name, searchinode, ra->count);
 
 /* . & .. */
+  if(debug > 1)
+    printf("\tADDENTRY:|%d/%d|\n", inode->inode,getpinode(inode));
   addentry(ra, &cp, &searchinode, inode->inode, ".");
   addentry(ra, &cp, &searchinode, getpinode(inode), "..");
 
@@ -490,9 +503,11 @@ nfsproc_setattr_2(sa)
   sattrargs *sa;
 {
   static struct attrstat res;
-  p_inode *inode = get_num(fh2inode(sa->file.data));
+  p_inode *inode;
   fattr *fp;
 
+  if(debug > 2) printf("setattr\n");
+  inode = get_num(fh2inode(sa->file.data));
   if(!inode)
     {
       if(debug) printf("setattr: stale fh\n");
@@ -534,11 +549,13 @@ remove_it(da, isdir)
   int isdir;
 {
   static nfsstat res;
-  p_inode *inode = get_num(fh2inode(da->dir.data));
-
+  p_inode *inode;
+  
+  if(debug > 2) printf("remove_it\n");
+  inode = get_num(fh2inode(da->dir.data));
   if(!inode)
     {
-      if(debug) printf("setattr: stale fh\n");
+      if(debug) printf("remove_it: stale fh\n");
       res = NO_PSION;
       return &res;
     }
@@ -574,10 +591,12 @@ nfsproc_rename_2(ra)
   renameargs *ra;
 {
   static nfsstat res;
-  p_inode *from = get_num(fh2inode(ra->from.dir.data)),
-          *to   = get_num(fh2inode(ra->to.dir.data));
+  p_inode *from, *to;
   char ldata[300], *old, c;
 
+  if(debug > 2) printf("rename\n");
+  from = get_num(fh2inode(ra->from.dir.data)),
+  to   = get_num(fh2inode(ra->to.dir.data));
   if(!from || !to)
     {
       if(debug) printf("rename: stale fh\n");
@@ -612,7 +631,7 @@ nfsproc_rename_2(ra)
 /* ARGSUSED */
 struct statfsres *
 nfsproc_statfs_2(fh)
-  struct nfs_fh *fh;
+  struct mynfs_fh *fh;
 {
   static statfsres res;
   statfsokres *rp;
@@ -651,14 +670,20 @@ nfsproc_statfs_2(fh)
  * Cache the read data. This cache will be invalidated if there are
  * no more requests in the queue for at least XXX seconds.
  * See also write :-(
+ *
+ *
+ * Another problem: since linux appears to drop bytes (no matter if using
+ * HW-flowcontrol or not), we retry our read request...
  */
+
 struct readres *
 nfsproc_read_2(ra)
   struct readargs *ra;
 {
+  static int read_retries = 0;
   static struct readres res;
   static unsigned char rop[NFS_MAXDATA];
-  p_inode *inode = get_num(fh2inode(ra->file.data));
+  p_inode *inode;
   fattr *fp;
   struct cache *cp;
   struct dcache *dcp;
@@ -666,6 +691,8 @@ nfsproc_read_2(ra)
   unsigned int  i_crc;
   unsigned char c_crc[2];
 
+  if(debug > 2) printf("read\n");
+  inode = get_num(fh2inode(ra->file.data));
   if(!inode)
     {
       if(debug) printf("read: stale fh\n");
@@ -712,9 +739,17 @@ nfsproc_read_2(ra)
 
   if(getcount(rop, len))
     {
+      if(read_retries++ < 2)
+        {
+	  if(debug)
+	    printf("Read: bad data received, retrying (%d)\n", read_retries);
+          return nfsproc_read_2(ra);
+	}
+      read_retries = 0;
       res.status = NO_PSION;
       return &res;
     }
+  read_retries = 0;
 
   if(!old_nfsc)
     {
@@ -832,7 +867,7 @@ nfsproc_write_2(wa)
   writeargs *wa;
 {
   static struct attrstat res;
-  p_inode *inode = get_num(fh2inode(wa->file.data));
+  p_inode *inode;
   struct cache *cp;
   struct dcache *dcp;
   fattr *fp;
@@ -840,6 +875,8 @@ nfsproc_write_2(wa)
   struct attrstat *gres;
   int c,len, dlen, doff;
 
+  inode = get_num(fh2inode(wa->file.data));
+  if(debug > 2) printf("write\n");
   if(!inode)
     {
       if(debug) printf("write: stale fh\n");
@@ -855,7 +892,7 @@ nfsproc_write_2(wa)
   /* fetch attributes */
   if ((cp = search_cache(attrcache, inode->inode)) == 0)
     {
-      gres = nfsproc_getattr_2((struct nfs_fh *)wa->file.data);
+      gres = nfsproc_getattr_2((struct mynfs_fh *)wa->file.data);
       if(gres->status != NFS_OK)
         {
 	  res.status = gres->status;
@@ -981,7 +1018,7 @@ nfsproc_link_2(la)
 /* ARGSUSED */
 struct readlinkres *
 nfsproc_readlink_2(fh)
-  struct nfs_fh *fh;
+  struct mynfs_fh *fh;
 {
   static readlinkres res;
 
