@@ -44,6 +44,11 @@
  * - status (0 == OK)
  * - data
  *
+ * If soft_flow is set, replace following character:
+ * 0x01     -> 0x01,0x01
+ * 0x11(^Q) -> 0x01,0x02
+ * 0x13(^S) -> 0x01,0x03
+ *
  */
 
 #include <OSdefs.h>
@@ -68,8 +73,13 @@
 
 #include <sys/errno.h>
 extern int errno;
+#if defined(HAVE_STRERROR)
+# define ERROR(x) strerror(x)
+#else
 #if !defined(__NetBSD__) && !defined(__GLIBC__) && !defined(__FreeBSD__)
-extern char *sys_errlist[];
+  extern char *sys_errlist[];
+#endif
+# define ERROR(x) sys_errlist[x]
 #endif
 
 void
@@ -110,6 +120,8 @@ senddata(p, len)
 {
   int ret;
   int lerr;
+  char *sbuf = 0;
+  extern int tcp;
 
   if(debug) printf("\t-> psion\n");
   psion_alive = fd_is_still_alive(psionfd,1);
@@ -119,12 +131,41 @@ senddata(p, len)
       return EOF;
     }
 
+  /* Rewrite ^S & ^Q characters */
+  if(soft_flow)
+    {
+      int i, j;
+      sbuf = malloc(len*2);
+      for(i = j = 0; i < len; i++, j++)
+	{
+	  if(p[i] == SOFTFLOW_PREFIX)
+	    {
+	      sbuf[j++] = SOFTFLOW_PREFIX;
+	      sbuf[j] = SOFTFLOW_PREFIX;
+	    }
+	  else if(p[i] == SOFTFLOW_RSTART)
+	    {
+	      sbuf[j++] = SOFTFLOW_PREFIX;
+	      sbuf[j] = SOFTFLOW_START;
+	    }
+	  else if(p[i] == SOFTFLOW_RSTOP)
+	    {
+	      sbuf[j++] = SOFTFLOW_PREFIX;
+	      sbuf[j] = SOFTFLOW_STOP;
+	    }
+	  else
+	    sbuf[j] = p[i];
+	}
+      p = sbuf; len = j;
+    }
+
+
   lerr = 0;
   if(debug > 1) printf("\tSending: %#x, %d\n", (unsigned int)p, len);
   while((ret = write(psionfd, p, len)) < len)
     {
       if(debug > 1)
-	fprintf(stderr, "\t\twrite(%d, %#x, %d) = %d\n",
+	printf("\t\twrite(%d, %#x, %d) = %d\n",
 	        psionfd, (unsigned int)p, len, ret);
       if(ret > 0)
 	{
@@ -136,6 +177,7 @@ senddata(p, len)
       if(ret < 0 && errno != EAGAIN)
         {
 	  perror("p3nfsd: Write");
+	  if(sbuf) free(sbuf);
 	  return 1;
 	}
 
@@ -143,14 +185,15 @@ senddata(p, len)
       if(lerr > 1)
 	{
 	  if(debug)
-	    fprintf(stderr, "\twrite=%d after flush/select\n", ret);
+	    printf("\twrite=%d after flush/select\n", ret);
+	  if(sbuf) free(sbuf);
 	  return 1;
 	}
 
 #ifdef NO_WRITE_SELECT
 
       if(debug > 1)
-	fprintf(stderr, "\t\tdraining serial line\n");
+	printf("\t\tdraining serial line\n");
       if(tcdrain(psionfd))
 	perror("p3nfsd: tcdrain");
 
@@ -165,7 +208,8 @@ senddata(p, len)
 	if((ret = select(psionfd+1, 0, &writefd, 0, &to)) <= 0)
 	  {
 	    if(debug)
-	      fprintf(stderr, "p3nfsd: writeselect returned %d\n", ret);
+	      printf("p3nfsd: writeselect returned %d\n", ret);
+	    if(sbuf) free(sbuf);
 	    return 1; 
 	  }
       }
@@ -173,7 +217,7 @@ senddata(p, len)
 #endif
     }
   if(debug > 1)
-    fprintf(stderr, "\t\twrite(%d, %#x, %d) = %d\n", psionfd, (unsigned int)p, len, ret);
+    printf("\t\twrite(%d, %#x, %d) = %d\n", psionfd, (unsigned int)p, len, ret);
 
 #if !defined(linux)
   /*
@@ -181,17 +225,18 @@ senddata(p, len)
    * ioctl(3, TCSBRK, 0x1) = 0
    */
 #endif
-  if(tcdrain(psionfd))
+  if(!tcp && tcdrain(psionfd))
     perror("p3nfsd: tcdrain");
+  if(sbuf) free(sbuf);
   return 0;
 }
 
-#define TOUT 5
+int bytes_read, missing_bytes;
 
 static int
 getbyte()
 {
-  static unsigned char readbuf[256];
+  static unsigned char readbuf[1024];
   static int idx = 0, len = 0;
   int err;
 
@@ -199,19 +244,27 @@ getbyte()
     {
       for(;;)
         {
+          extern int timeout;
 #ifdef __SVR4
           struct pollfd readfd;
-
-	  errno = 0;
-	  readfd.fd = psionfd;
-	  readfd.events = POLLIN;
-	  err = poll(&readfd, 1, TOUT * 1000);
 #else
           fd_set readfd;
           struct timeval to;
+#endif
 
+	  if(debug > 1) { 
+	    printf("\t\tRead select (need %d, to %d)...",missing_bytes,timeout);
+	    fflush(stdout);
+	  }
+#ifdef __SVR4
 	  errno = 0;
-	  to.tv_sec = TOUT; to.tv_usec = 0; 
+	  readfd.fd = psionfd;
+	  readfd.events = POLLIN;
+	  err = poll(&readfd, 1, timeout);
+#else
+	  errno = 0;
+	  to.tv_sec = timeout/1000;
+	  to.tv_usec = (timeout-to.tv_sec*1000) * 1000;
 	  FD_ZERO(&readfd); FD_SET(psionfd, &readfd);
 	  err = select(psionfd+1, &readfd, 0, 0, &to);
 #endif
@@ -219,17 +272,37 @@ getbyte()
 	    {
 	      if(errno == EINTR)
 	        continue;
-	      fprintf(stderr, "p3nfsd: select: %s\n",
-	              err < 0 ? sys_errlist[errno] : "timeout" );
+	      if(debug)
+		printf("select: %s\n", err<0 ? ERROR(errno) : "timeout");
 	      return EOF;
 	    }
 	  break;
 	}
 
       idx = 0;
-      if(debug > 1) { printf("\t\tFilling read buffer..."); fflush(stdout); }
       len = read(psionfd, readbuf, sizeof(readbuf));
       if(debug > 1) printf("got %d bytes\n", len);
+      if(soft_flow)
+        {
+	  int i, j;
+	  for(j = i = 0; i < len; i++, j++)
+	    {
+	      if(readbuf[i] == SOFTFLOW_PREFIX)
+	        {
+		  i++;
+	          if(readbuf[i] == SOFTFLOW_PREFIX)
+	            readbuf[j] = SOFTFLOW_PREFIX;
+	          else if(readbuf[i] == SOFTFLOW_START)
+	            readbuf[j] = SOFTFLOW_RSTART;
+	          else if(readbuf[i] == SOFTFLOW_STOP)
+	            readbuf[j] = SOFTFLOW_RSTOP;
+		}
+	      else
+	        readbuf[j] = readbuf[i];
+	    }
+	  len = j;
+          if(debug > 1) printf("Softflow got %d bytes\n", len);
+	}
       if(debug > 2) 
         {
 	  int i;
@@ -242,9 +315,12 @@ getbyte()
 	    }
 	  printf("\n");
 	}
+
       if(len <= 0)
         {
-	  perror("p3nfsd: read"); len = 0;
+	  fprintf(stderr, "p3nfsd: read len: %d, errno: %d/", len, errno);
+	  perror("");
+	  len = 0;
 	  return EOF;
 	}
     }
@@ -289,6 +365,7 @@ sendcmd(cmd, fname, rest, restlen)
   return senddata((char *)mbuf, l + (old_nfsc ? 0 : 1));
 }
 
+
 int
 getcount(str, num)
   unsigned char *str; int num;
@@ -296,14 +373,20 @@ getcount(str, num)
   int c;
 
   if(debug > 1) {printf("\tGoing to read %d bytes...",num);fflush(stdout);}
+  bytes_read = 0;
+  missing_bytes = num;
   while(num-- > 0)
-    if((c = getbyte()) == EOF)
-      {
-	if(debug) printf("Bad connection to psion... (%d)\n", c);
-	return -53;
-      }
-    else
-      *str++ = c;
+    {
+      if((c = getbyte()) == EOF)
+	{
+	  if(debug) printf("Bad connection to psion... (%d)\n", c);
+	  return -53;
+	}
+      else
+	*str++ = c;
+      bytes_read++;
+      missing_bytes--;
+    }
   if(debug > 1) printf("\tRead ok\n");
   return 0;
 }
@@ -316,6 +399,8 @@ getstr(str)
   unsigned char *s = (unsigned char *)str;
 
   if(debug > 1) {printf("\tGoing to read string...");fflush(stdout);}
+  bytes_read = 0;
+  missing_bytes = 1;
   for(;;)
     {
       if((c = getbyte()) == EOF)
@@ -338,6 +423,8 @@ getanswer()
 {
   int i;
 
+  bytes_read = 0;
+  missing_bytes = 1;
   i = getbyte();
   if(debug > 1) printf("Answer:(1) %#x\n", i);
   if(old_nfsc)
